use std::collections::{HashMap, HashSet};
use std::process::Command;
use std::str::FromStr;

use ccutils::cmd::IsOk;
use cvrf_xmlparser::{Severity, CUSA};
use lazy_static::lazy_static;
use rpm_rs::rpm::{get_installed_packages, Package};
use rpm_rs::rpmio::rpmvercmp;
use updateinfo_xmlparser::{RpmInfo, UpdateInfoDb};

use crate::analyzer::db::{CveDb, PackgeDb, SaDb};
use crate::cli::Cli;
use crate::CONFIG;

lazy_static! {
    pub static ref update_pkgs: PackgeDb = {
        let mut updatedb = UpdateInfoDb::new();
        updatedb.load_xml(&CONFIG.updateinfo()).unwrap();

        let mut pkgdb = PackgeDb::new();
        pkgdb.load_from_updateinfodb(&updatedb);
        pkgdb
    };
    pub static ref sadb: SaDb = {
        let mut _sadb = SaDb::new();

        _sadb.load_from_file(&CONFIG.sainfo()).unwrap();
        _sadb
    };
    pub static ref cvedb: CveDb = sadb.get_cvedb();
}

pub mod db;

struct Reporter {
    // 已修复的安全更新
    fixed: HashMap<CUSA, HashSet<RpmInfo>>,

    // 可用的安全更新
    avaliable: HashMap<CUSA, HashSet<RpmInfo>>,
}

impl Reporter {
    // 新建一个空的 Reporter
    pub fn new() -> Self {
        let fixed = HashMap::new();
        let avaliable = HashMap::new();

        Self { fixed, avaliable }
    }

    // get fixed, 不可写
    pub fn get_fixed(&self) -> &HashMap<CUSA, HashSet<RpmInfo>> {
        &self.fixed
    }

    // get avaliable, 只读
    pub fn get_avaliable(&self) -> &HashMap<CUSA, HashSet<RpmInfo>> {
        &self.avaliable
    }

    // 添加包到已修复
    pub fn add_to_fixed(&mut self, cusa: &CUSA, rpm: RpmInfo) {
        // 可以找到则直接 insert
        if let Some(rpms) = self.fixed.get_mut(cusa) {
            rpms.insert(rpm);
        } else {
            let mut rpms = HashSet::new();
            rpms.insert(rpm);
            self.fixed.insert(cusa.clone(), rpms);
        }
    }

    // 添加包到可用更新
    pub fn add_to_avaliable(&mut self, cusa: &CUSA, rpm: RpmInfo) {
        // 可以找到则直接 insert
        if let Some(rpms) = self.avaliable.get_mut(cusa) {
            rpms.insert(rpm);
        } else {
            let mut rpms = HashSet::new();
            rpms.insert(rpm);
            self.avaliable.insert(cusa.clone(), rpms);
        }
    }
}

/// 这里定义了 RPM 包的版本信息的特征，
/// 其返回值的类型，取决于 librpm.rs 中 `Package` 数据类型的返回值。
/// 对于其所提供的函数，当前是够用即可的状态。
#[allow(dead_code)]
trait RpmVerCmp {
    fn get_name(&self) -> &str;

    fn get_epoch(&self) -> Option<i32>;

    fn get_version(&self) -> &str;

    fn get_release(&self) -> &str;

    fn get_arch(&self) -> Option<&str>;

    fn get_evr(&self) -> String;
}

// 为 Package 实现 RpmVerCmp 特征
impl RpmVerCmp for Package {
    fn get_name(&self) -> &str {
        self.name()
    }

    fn get_epoch(&self) -> Option<i32> {
        self.epoch()
    }

    fn get_version(&self) -> &str {
        self.version()
    }

    fn get_release(&self) -> &str {
        self.release()
    }

    fn get_arch(&self) -> Option<&str> {
        self.arch()
    }

    fn get_evr(&self) -> String {
        self.evr()
    }
}

// 为 RpmInfo 实现 RpmVerCmp 特征
impl RpmVerCmp for RpmInfo {
    fn get_name(&self) -> &str {
        self.name()
    }

    fn get_epoch(&self) -> Option<i32> {
        self.epoch()
    }

    fn get_version(&self) -> &str {
        self.version()
    }

    fn get_release(&self) -> &str {
        self.release()
    }

    fn get_arch(&self) -> Option<&str> {
        self.arch()
    }

    fn get_evr(&self) -> String {
        self.evr()
    }
}

pub fn cuvat_run(cli: &Cli) -> crate::Result<()> {
    // 只报告 cve 相关
    if cli.cves {
        return list_cves(cli);
    }

    // 只报告 sa 相关
    if cli.sas {
        return list_sas(cli);
    }

    // 生成报告
    if cli.report {
        return repoter(cli);
    }

    // 默认报告格式
    summary(cli)
}

fn list_cves(cli: &Cli) -> crate::Result<()> {
    if cli.dnf {
        return _dnf_list_cves(cli);
    }

    let reporter = _reporter()?;
    let avaliable = reporter.get_avaliable();
    let fixed = reporter.get_fixed();
    let severity = Severity::from_str(&cli.severity)?;
    let mut _cves = HashSet::new();
    cli.sources.iter().for_each(|id| {
        _cves.insert(id);
    });

    let mut cves = HashSet::new();
    let mut effected_rpms: HashMap<String, Vec<String>> = HashMap::new();
    let mut fixed_rpms: HashMap<String, Vec<String>> = HashMap::new();
    let mut msg = String::new();

    // 已修复的 CVE
    // 列出修复的软件包
    for (cusa, rpms) in fixed {
        // 过滤
        if cusa.severity() < &severity {
            continue;
        }

        cusa.cves().iter().for_each(|cve| {
            if let Some(_) = _cves.get(&cve.id) {
                let _ = cves.insert(cve.clone());
                if let Some(fixed) = fixed_rpms.get_mut(&cve.id) {
                    for rpm in rpms {
                        fixed.push(rpm.name().into());
                    }
                } else {
                    let fixed: Vec<String> = rpms.into_iter().map(|rpm| rpm.name().into()).collect();
                    fixed_rpms.insert(cve.id.clone(), fixed);
                }
                //_cves.remove(&cve.id);
            }
        });
    }

    // 仅针对可用更新
    for (cusa, rpms) in avaliable {
        // 过滤
        if cusa.severity() < &severity {
            continue;
        }

        cusa.cves().iter().for_each(|cve| {
            if let Some(_) = _cves.get(&cve.id) {
                let _ = cves.insert(cve.clone());
                if let Some(effected) = effected_rpms.get_mut(&cve.id) {
                    for rpm in rpms {
                        effected.push(rpm.name().into());
                    }
                } else {
                    let effected: Vec<String> = rpms.into_iter().map(|rpm| rpm.name().into()).collect();
                    effected_rpms.insert(cve.id.clone(), effected);
                }
                //_cves.remove(&cve.id);
            }
        });
    }

    for cve in &cves {
        if cli.info {
            msg = format!("{msg}\n\n{cve:#?}\n");
            if let Some(_fixed) = fixed_rpms.get(&cve.id) {
                msg = format!("{msg}已修复的软件： {:?}\n", _fixed);
            }
            if let Some(_effected) = effected_rpms.get(&cve.id) {
                msg = format!("{msg}可用的更新： {:?}\n", _effected);
            }
        } else {
            msg = format!("{msg}\n{}", cve.id);
        }
        // 重置此 cve
        _cves.remove(&cve.id);
    }

    if !_cves.is_empty() {
        msg = format!("{msg}\n\n未知的 CVEs（请联系支持人员）:\n{_cves:#?}")
    }

    println!("{msg}");

    Ok(())
}

fn _dnf_list_cves(_cli: &Cli) -> crate::Result<()> {
    Ok(())
}

fn list_sas(cli: &Cli) -> crate::Result<()> {
    if cli.dnf {
        return _dnf_list_sas(cli);
    }

    let reporter = _reporter()?;
    let avaliable = reporter.get_avaliable();
    let severity = Severity::from_str(&cli.severity)?;
    let mut _sas = HashSet::new();
    cli.sources.iter().for_each(|id| {
        _sas.insert(id);
    });

    let mut sas = HashSet::new();
    let mut msg = String::new();

    // 仅针对可用更新
    for (cusa, _) in avaliable {
        // 过滤
        if cusa.severity() < &severity {
            continue;
        }

        if let Some(_) = _sas.get(&cusa.id) {
            let _ = sas.insert(cusa.clone());
            _sas.remove(&cusa.id);
        }
    }

    for sa in &sas {
        if cli.info {
            msg = format!("{msg}\n\n{sa:#?}");
        } else {
            msg = format!("{msg}\n{}", sa.id);
        }
    }

    if !_sas.is_empty() {
        msg = format!("{msg}\n\n未知的安全公告（请联系支撑人员）：\n{_sas:#?}")
    }

    println!("{msg}");

    Ok(())
}

fn _dnf_list_sas(_cli: &Cli) -> crate::Result<()> {
    Ok(())
}

// 最为详尽的报告，包括当前系统中软件包所涉及到，已修复和可用但未修复
// 的所有软件包的列表。
// 其格式如下：
// XXXXXX 报告
// FIXED:
// SA ID
//   CVES
//     effected rpms list
// ...
//
// ====================
// AVALIABLE:
// SA ID
//   CVES
//     effected rpms list
// ...
//
fn repoter(cli: &Cli) -> crate::Result<()> {
    if cli.dnf {
        println!("暂不支持此功能，请联系开发者");
        return Ok(());
    }

    let reporter = _reporter()?;
    let fixed = reporter.get_fixed();
    let avaliable = reporter.get_avaliable();
    let mut msg = String::new();

    msg = format!("{msg}系统漏洞信息检测报告\n\n");

    // fixed
    msg = format!("{msg}已修复的安全公告\n");
    for (cusa, rpms) in fixed {
        // sa cves 信息
        msg = format!("{msg}{} (fixed):\n", cusa.id());
        let mut cves = vec![];
        cusa.cves().iter().for_each(|cve| {
            cves.push(&cve.id);
        });
        msg = format!("{msg}{:2}涉及 CVE 列表：{:?}\n", "", cves);

        for rpm in rpms {
            msg = format!("{msg}{:4}{}\n", "", rpm.nevra());
        }
        // 空
        msg = format!("{msg}\n");
    }

    // "==============="
    let mut seprator = String::new();
    for _ in 0..80 {
        seprator.push('=');
    }
    msg = format!("{msg}\n{seprator}\n\n\n");

    // avaliable updates
    msg = format!("{msg}可用更新\n");
    for (cusa, rpms) in avaliable {
        // sa cves 信息
        msg = format!("{msg}{} (avaliable):\n", cusa.id());
        let mut cves = vec![];
        cusa.cves().iter().for_each(|cve| {
            cves.push(&cve.id);
        });
        msg = format!("{msg}{:2}涉及 CVE 列表：{:?}\n", "", cves);

        for rpm in rpms {
            msg = format!("{msg}{:4}{}\n", "", rpm.nevra());
        }
        // 空
        msg = format!("{msg}\n");
    }

    println!("{msg}");

    // summary 一下
    summary(cli)
}

fn summary(cli: &Cli) -> crate::Result<()> {
    if cli.dnf {
        return _dnf_summary(cli);
    }

    let reporter = _reporter()?;
    let avaliable = reporter.get_avaliable();
    let severity = Severity::from_str(&cli.severity)?;
    let mut total = 0;
    let mut res = vec![0; 5];
    let mut lists = vec![vec![]; 5];
    let mut msg = String::new();

    for (cusa, _) in avaliable {
        // 过滤
        if cusa.severity() < &severity {
            continue;
        }

        let pos: usize = cusa.severity().into();
        total += 1;
        res[pos] += 1;
        lists[pos].push(cusa);
    }

    // 如果需要列出详细的 sa
    if cli.list {
        msg = format!("当前系统未修复的公告列表（共 {total} 个）：\n");
        for sas in &lists {
            for sa in sas {
                msg = format!("{msg}\n{:>4}{}: {:?}", "", sa.id(), sa.severity());
            }
        }
        //msg = format!("{msg}\n\n");
    } else {
        msg = format!(
            "{msg}
本次安全检查共扫描出 {} 个漏洞：

    致命漏洞：{:>3} 个
    高危漏洞：{:>3} 个
    中危漏洞：{:>3} 个
    低危漏洞：{:>3} 个
    ",
            total, res[4], res[3], res[2], res[1]
        );
    }

    println!("{msg}");

    Ok(())
}

fn _dnf_summary(_cli: &Cli) -> crate::Result<()> {
    let cmd = Command::new("dnf")
        .arg("updateinfo")
        .output()
        .expect("执行 `dnf udpateinfo` 命令失败");

    match cmd.result() {
        Ok(res) => println!("{res}"),
        Err(err) => eprintln!("{err}"),
    }

    Ok(())
}

// TODO: 也可从文件中读取某机的已安装的软件列表，进行分析
fn _reporter() -> crate::Result<Reporter> {
    let mut reporter = Reporter::new();
    // 当前系统所有已安装的 rpm 包
    let installed = get_installed_packages();
    // installed 已被消费掉
    // latest_installed 是所有最新版本软件包的 Vec
    let latest_installed: Vec<RpmInfo> = installed
        .into_iter()
        .map(|(_, pkgs)| {
            let mut latest = pkgs[0].clone();
            for pkg in pkgs.into_iter() {
                latest = rpmdb_package_vercmp(latest, pkg);
            }
            // 转换成 RpmInfo ，统一后方便后续文件分析支持
            package2rpminfo(latest)
        })
        .collect();

    // 获取 sa 更新列表
    for pkg in latest_installed {
        let name = pkg.name();
        if let Some(updates) = update_pkgs.db().get(name) {
            for update in updates {
                // 因 updateinfo 里包含所有架构，故于此做个判断
                // 是否考虑架构信息在不同的版本有所不同？
                match (update.arch(), pkg.arch()) {
                    (Some(_a), Some(__a)) => {
                        if _a != __a {
                            continue;
                        }
                    },
                    _ => continue
                }
                /*
                if update.arch() != pkg.arch().unwrap() {
                    continue;
                }
                */
                let sainfo = sadb.db().get(update.sa()).unwrap();
                // epoch 判断
                match (update.epoch(), pkg.epoch()) {
                    (Some(_), None) => {
                        reporter.add_to_avaliable(&sainfo, update.clone());
                    }
                    (None, Some(_)) => {
                        reporter.add_to_fixed(&sainfo, update.clone());
                    }
                    _ => {}
                }

                // evr 对比
                if rpmvercmp(&update.evr(), &pkg.evr()) > 0 {
                    reporter.add_to_avaliable(&sainfo, update.clone());
                } else {
                    reporter.add_to_fixed(&sainfo, update.clone());
                }
            }
        }
    }

    Ok(reporter)
}

// 对比两个 rpm Package 的版本，返回最新的一个
fn rpmdb_package_vercmp<P: RpmVerCmp>(pa: P, pb: P) -> P {
    // 首先进行 epoch 的比较
    match (pa.get_epoch(), pb.get_epoch()) {
        (Some(_), None) => return pa,
        (None, Some(_)) => return pb,
        _ => {} // 继续往下对比
    }

    if rpmvercmp(&pa.get_evr(), &pb.get_evr()) > 0 {
        return pa;
    } else {
        return pb;
    }
}

// 将 RPMDB 里的 Package 转换为 RpmInfo，方便统一进行版本对比和处理。
fn package2rpminfo(pkg: Package) -> RpmInfo {
    let mut rpminfo = RpmInfo::new();

    // 目前只需要将 name/epoch/version/release/arch 几项填充即可
    rpminfo.set("name", pkg.name().to_string());
    // Option<i32> => String
    if let Some(epoch) = pkg.epoch() {
        rpminfo.set("epoch", epoch.to_string());
    }
    rpminfo.set("version", pkg.version().to_string());
    rpminfo.set("release", pkg.release().to_string());
    // Option<&str> => String
    if let Some(arch) = pkg.arch() {
        rpminfo.set("arch", arch.to_string());
    }

    rpminfo
}
